# Copyright (c) 2019 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from bark.world.agent import Agent
from bark.world import World
from bark.world.map import MapInterface
from modules.runtime.commons.parameters import ParameterServer
from modules.runtime.commons.xodr_parser import XodrParser
import copy


class Scenario:
  def __init__(self,
               agent_list=None,
               eval_agent_ids=None,
               map_file_name=None,
               json_params=None,
               map_interface=None):
    self._agent_list = agent_list or []
    self._eval_agent_ids = eval_agent_ids or []
    self._map_file_name = map_file_name
    self._json_params = json_params
    self._map_interface = map_interface

  @property
  def map_file_name(self):
    return self._map_file_name
  
  @property
  def json_params(self):
    return self._json_params

  @property
  def eval_agent_ids(self):
    return self._eval_agent_ids

  def get_world_state(self):
    """get initial world state of scenario to start simulation from here
    
    Returns:
        [bark.world.World]
    """
    return self._build_world_state()

  def copy(self):
    return Scenario(agent_list=copy.deepcopy(self._agent_list),
                    eval_agent_ids=self._eval_agent_ids.copy(),
                    map_file_name=self._map_file_name,
                    json_params=self._json_params.copy(),
                    map_interface=self._map_interface)

  def _build_world_state(self):
    param_server = ParameterServer(json=self._json_params)
    world = World(param_server)
    if self._map_interface is None:
      world = self.setup_map(world, self._map_file_name)
    else:
      world.SetMap(self._map_interface)
    for agent in self._agent_list:
      agent.GenerateRoadCorridor(self._map_interface)
      world.AddAgent(agent)
    return world

  def __getstate__(self):
    odict = self.__dict__.copy()
    del odict['_map_interface']
    return odict

  def __setstate__(self, sdict):
    sdict['_map_interface'] = None
    self.__dict__.update(sdict)

  def setup_map(self, world, _map_file_name):
    if not _map_file_name:
        return world
    xodr_parser = XodrParser(_map_file_name )
    map_interface = MapInterface()
    map_interface.SetOpenDriveMap(xodr_parser.map)
    self._map_interface = map_interface
    world.SetMap(map_interface)
    return world








# Copyright (c) 2019 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import time

def return_execution_time(func):
  def wrapper(*args, **kwargs):
    time_before_step = time.time()
    func(*args, **kwargs)
    step_duration = time.time() - time_before_step
    return step_duration

  return wrapper


# Copyright (c) 2019 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import pygame as pg
import numpy as np
from bark.geometry import *
from bark.viewer import *
from bark.world.opendrive import *
from bark.models.dynamic import *
from modules.runtime.viewer.viewer import BaseViewer


class PygameViewer(BaseViewer):
    def __init__(self, params=None, **kwargs):
        super(PygameViewer, self).__init__(params=params, **kwargs)

        self.screen_dims = kwargs.pop("screen_dims", np.array([1024, 1024]))
        self.screen_width = self.screen_dims[0]
        self.screen_height = self.screen_dims[1]

        self.screen_map_ratio = max(self.screen_width/(np.diff(self.world_x_range)[0]),
                                    self.screen_height/(np.diff(self.world_y_range)[0]))
        self.camera_view_size = np.array([np.diff(self.dynamic_world_x_range)[0], np.diff(self.dynamic_world_y_range)[0]])

        self.map_surface = None
        self.map_size = None

        self.screen_surface = pg.Surface((self.screen_width, self.screen_height))

        pg.font.init()

        try:
            self.screen = pg.display.set_mode((self.screen_width, self.screen_height), pg.DOUBLEBUF)
            pg.display.set_caption("Bark")

            self.clear()
            self.show()
        except pg.error:
            self.screen = None
            print("No available video device")

    def drawMap(self, map):
        if self.map_surface == None:
            # find boundary to create fixed size pygame surface
            self.map_min_boundary = np.full(2, np.inf)
            self.map_max_boundary = np.full(2, -np.inf)
            lanes_np = []
            lanes_dashed = []

            for _, road in map.GetRoads().items():
                for lane_section in road.lane_sections:
                    for _, lane in lane_section.GetLanes().items():
                        lane_np = lane.line.ToArray()
                        lanes_np.append(lane_np)
                        lanes_dashed.append(lane.road_mark.type == XodrRoadMarkType.broken
                                            or lane.road_mark.type == XodrRoadMarkType.none)

                        self.map_min_boundary = np.minimum(self.map_min_boundary, np.amin(lane_np, axis=0))
                        self.map_max_boundary = np.maximum(self.map_max_boundary, np.amax(lane_np, axis=0))

            self.map_size = self.map_max_boundary-self.map_min_boundary
            # the size needed to be scaling larger for better visualization
            # as a pygame surface is stored as pixels
            self.map_surface = pg.Surface(tuple(self.map_size*self.screen_map_ratio))
            self.map_surface.fill((255, 255, 255))

            for lane_np, lane_dashed in zip(lanes_np, lanes_dashed):
                pg.draw.aalines(self.map_surface, self.getColor(self.color_lane_boundaries), False, self.mapToSurfaceCoordinates(lane_np), 3)

        camera_coordinate = self.mapToSurfaceCoordinates(np.array([self.dynamic_world_x_range[0], self.dynamic_world_y_range[1]]))
        camera_view_range = np.array([np.diff(self.dynamic_world_x_range)[0], np.diff(self.dynamic_world_y_range)[0]])*self.screen_map_ratio
        self.screen_surface.blit(self.map_surface, (0, 0), tuple(np.around(np.concatenate((camera_coordinate, camera_view_range)))))

    def drawPoint2d(self, point2d, color, alpha):
        pg.draw.circle(self.screen_surface, self.getColor(color), self.pointsToCameraCoordinate(point2d), 1, 0)

    def drawLine2d(self, line2d, color='blue', alpha=1.0, dashed=False):
        # TODO: enable dashed line
        line2d = self.pointsToCameraCoordinate(line2d)
        pg.draw.lines(self.screen_surface, self.getColor(color), False, line2d, 3)

    def drawPolygon2d(self, polygon, color, alpha):
        points = self.pointsToCameraCoordinate(polygon)
        pg.draw.polygon(self.screen_surface, self.getColor(color), points)

    def drawTrajectory(self, trajectory, color):
        if len(trajectory) < 1:
            return
        point_list = []
        for state in trajectory:
            point_list.append([state[round(StateDefinition.X_POSITION)], state[round(StateDefinition.Y_POSITION)]])

        pg.draw.lines(self.screen_surface, self.getColor(color), False, self.pointsToCameraCoordinate(point_list), 5)

    def drawText(self, position, text, **kwargs):
        font = pg.font.get_default_font()
        fontsize = kwargs.pop("fontsize", 18)
        color = kwargs.pop("color", (0, 0, 0))
        background_color = kwargs.pop("background_color", (255, 255, 255))
        text_surface = pg.font.SysFont(font, fontsize).render(text, True, color, background_color)
        self.screen_surface.blit(text_surface, (position[0] * self.screen_width, position[1] * self.screen_height))

    def getColor(self, color):
        if isinstance(color, Viewer.Color):
            return {
                Viewer.Color.White: (255, 255, 255),
                Viewer.Color.Red: (255, 0, 0),
                Viewer.Color.Blue: (0, 0, 255),
                Viewer.Color.Magenta: (100, 255, 100),
                Viewer.Color.Brown: (150, 50, 50),
                Viewer.Color.Black: (0, 0, 0)
            }.get(color, (0, 0, 0))
        else:
            return (color[0] * 255, color[1] * 255, color[2] * 255)

    def drawWorld(self, world, eval_agent_ids=None, show=True):
        self.clear()
        super(PygameViewer, self).drawWorld(world, eval_agent_ids)
        if show:
            self.show()

    def show(self, block=True):
        if self.screen is None:
            return
        self.screen.blit(self.screen_surface, (0, 0))
        pg.display.flip()
        pg.event.get()  # call necessary for visbility of pygame viewer on macos

    def clear(self):
        self.screen_surface.fill((255, 255, 255))

    """
        points: numpy array
        return: numpy array

        The origin of pygame surface is located at top left, increment downward
        therefore all the coordinates need to be transformed

        1. Mirror by y-axis
        2. Translate y-axis by the max y-coordinate in the map
        3. Translate x-axis by the max x-coordinate in the map
        4. Scaling by screen_map_ratio
    """

    def mapToSurfaceCoordinates(self, points):
        return (points * np.array([1, -1]) + np.array([-self.map_min_boundary[0], self.map_max_boundary[1]])) * self.screen_map_ratio

    def pointsToCameraCoordinate(self, points):
        if isinstance(points, list):
            points = np.array(points)
        elif not isinstance(points, np.ndarray):
            points = points.ToArray()

        return np.array([0, self.screen_height])+(points - np.array([self.dynamic_world_x_range[0], self.dynamic_world_y_range[0]]))*np.array([1, -1]) \
            / self.camera_view_size*self.screen_dims



# Copyright (c) 2019 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import numpy as np
import time
import os
import bark_project
from bark_project.modules.runtime.commons.parameters import ParameterServer
from bark_project.modules.runtime.viewer.matplotlib_viewer import MPViewer
from bark_project.modules.runtime.commons.xodr_parser import XodrParser
from bark.models.behavior import BehaviorConstantVelocity
from bark.models.execution import ExecutionModelInterpolate
from bark.models.dynamic import SingleTrackModel
from bark.world import World
from bark.world.goal_definition import GoalDefinitionPolygon
from bark.world.agent import Agent
from bark.world.map import MapInterface
from bark.geometry.standard_shapes import CarLimousine
from bark.geometry import Point2d, Polygon2d

def run_example():

    # Parameters Definitions
    param_server = ParameterServer(filename=os.path.join(os.path.dirname(__file__), "params/od8_const_vel_one_agent.json"))
    # set parameter that is accessible in Python as well as cpp
    # param_server.setReal("wheel_base", 0.8)

    # World Definition
    world = World(param_server)

    # Model Definitions
    behavior_model = BehaviorConstantVelocity(param_server)
    execution_model = ExecutionModelInterpolate(param_server)
    dynamic_model = SingleTrackModel(param_server)

    # Map Definition
    xodr_parser = XodrParser(os.path.join(os.path.dirname(__file__), "../modules/runtime/tests/data/Crossing8Course.xodr"))
    map_interface = MapInterface()
    map_interface.SetOpenDriveMap(xodr_parser.map)
    world.SetMap(map_interface)

    # Agent Definition
    agent_2d_shape = CarLimousine()
    init_state = np.array([0, -15, -13, 3.14*5.0/4.0, 10/3.6])
    agent_params = param_server.addChild("agent1")
    goal_polygon = Polygon2d([0, 0, 0],[Point2d(-1,-1),Point2d(-1,1),Point2d(1,1), Point2d(1,-1)])
    goal_polygon = goal_polygon.Translate(Point2d(-191.789,-50.1725))

    agent = Agent(init_state,
                  behavior_model,
                  dynamic_model,
                  execution_model,
                  agent_2d_shape,
                  agent_params,
                  GoalDefinitionPolygon(goal_polygon), # goal_lane_id
                  map_interface)
    world.AddAgent(agent)

    # viewer
    viewer = MPViewer(params=param_server,
                      use_world_bounds=True)

    # World Simulation
    sim_step_time = param_server["simulation"]["step_time",
                                               "Step-time in simulation",
                                               0.05]
    sim_real_time_factor = param_server["simulation"]["real_time_factor",
                                                      "execution in real-time or faster",
                                                      100]

    for _ in range(0, 10):
      viewer.clear()
      world.Step(sim_step_time)
      viewer.drawWorld(world)
      viewer.drawRoadCorridor(agent.road_corridor)
      viewer.show(block=False)
      time.sleep(sim_step_time/sim_real_time_factor)

    param_server.save(s.path.join(os.path.dirname(__file__), "params/od8_const_vel_one_agent_written.json"))
